from typing import Dict

import pandas as pd

from sec_cik_mapper import MutualFundMapper


def test_mutual_fund_mapper_initialization(mutual_fund_mapper: MutualFundMapper):
    df = mutual_fund_mapper.raw_dataframe

    # Verify dataframe rows
    assert len(df) > 0

    # Verify dataframe columns
    df_columns = list(df.columns)
    assert len(df_columns) == 4
    assert "CIK" in df_columns
    assert "Ticker" in df_columns
    assert "Series ID" in df_columns
    assert "Class ID" in df_columns

    validate_dataframe_content(df)


def validate_dataframe_content(df: pd.DataFrame):
    for cik, ticker, seriesId, classId in zip(
        df["CIK"], df["Ticker"], df["Series ID"], df["Class ID"]
    ):
        assert isinstance(cik, str) and len(cik) == 10
        # SEC outputs blank tickers for some mutual funds
        assert isinstance(ticker, str)
        assert isinstance(seriesId, str) and len(seriesId) > 0
        assert isinstance(classId, str) and len(classId) > 0


def test_cik_to_tickers(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    cik_to_tickers = mutual_fund_mapper.cik_to_tickers
    df = mutual_fund_mapper.raw_dataframe
    assert len(cik_to_tickers) == df[df.Ticker != ""].CIK.nunique()

    # Deal with CIKs mapping to multiple tickers
    num_tickers = 0
    for cik, tickers in cik_to_tickers.items():
        assert isinstance(tickers, set)
        num_tickers += len(tickers)
        assert len(cik) == 10

    df = mutual_fund_mapper.raw_dataframe.Ticker
    assert num_tickers == df[df != ""].nunique()

    ticker = vtsax_mutual_fund["ticker"]
    cik = vtsax_mutual_fund["cik"]
    assert cik in cik_to_tickers
    assert ticker in cik_to_tickers[cik]


def test_ticker_to_cik(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    ticker_to_cik = mutual_fund_mapper.ticker_to_cik
    df = mutual_fund_mapper.raw_dataframe.Ticker
    assert len(ticker_to_cik) == df[df != ""].nunique()

    ticker = vtsax_mutual_fund["ticker"]
    cik = vtsax_mutual_fund["cik"]
    assert ticker in ticker_to_cik
    assert len(ticker_to_cik[ticker]) == 10
    assert ticker_to_cik[ticker] == cik


def test_cik_to_series_ids(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    cik_to_series_ids = mutual_fund_mapper.cik_to_series_ids
    assert len(cik_to_series_ids) == mutual_fund_mapper.raw_dataframe.CIK.nunique()

    series_id = vtsax_mutual_fund["seriesId"]
    cik = vtsax_mutual_fund["cik"]
    assert cik in cik_to_series_ids
    assert series_id in cik_to_series_ids[cik]


def test_ticker_to_series_id(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    ticker_to_series_id = mutual_fund_mapper.ticker_to_series_id
    df = mutual_fund_mapper.raw_dataframe.Ticker
    assert len(ticker_to_series_id) == df[df != ""].nunique()

    ticker = vtsax_mutual_fund["ticker"]
    series_id = vtsax_mutual_fund["seriesId"]
    assert ticker in ticker_to_series_id
    assert ticker_to_series_id[ticker] == series_id


def test_series_id_to_cik(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    series_id_to_cik = mutual_fund_mapper.series_id_to_cik
    assert (
        len(series_id_to_cik) == mutual_fund_mapper.raw_dataframe["Series ID"].nunique()
    )

    cik = vtsax_mutual_fund["cik"]
    series_id = vtsax_mutual_fund["seriesId"]
    assert series_id in series_id_to_cik
    assert series_id_to_cik[series_id] == cik


def test_series_id_to_tickers(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    series_id_to_tickers = mutual_fund_mapper.series_id_to_tickers
    df = mutual_fund_mapper.raw_dataframe
    ticker_df = mutual_fund_mapper.raw_dataframe["Ticker"]
    assert len(series_id_to_tickers) == df[ticker_df != ""]["Series ID"].nunique()

    ticker = vtsax_mutual_fund["ticker"]
    series_id = vtsax_mutual_fund["seriesId"]
    assert series_id in series_id_to_tickers
    assert ticker in series_id_to_tickers[series_id]


def test_series_id_to_class_ids(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    series_id_to_class_ids = mutual_fund_mapper.series_id_to_class_ids
    assert (
        len(series_id_to_class_ids)
        == mutual_fund_mapper.raw_dataframe["Series ID"].nunique()
    )

    class_id = vtsax_mutual_fund["classId"]
    series_id = vtsax_mutual_fund["seriesId"]
    assert series_id in series_id_to_class_ids
    assert class_id in series_id_to_class_ids[series_id]


def test_ticker_to_class_id(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    ticker_to_class_id = mutual_fund_mapper.ticker_to_class_id
    df = mutual_fund_mapper.raw_dataframe.Ticker
    assert len(ticker_to_class_id) == df[df != ""].nunique()

    class_id = vtsax_mutual_fund["classId"]
    ticker = vtsax_mutual_fund["ticker"]
    assert ticker in ticker_to_class_id
    assert ticker_to_class_id[ticker] == class_id


def test_cik_to_class_ids(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    cik_to_class_ids = mutual_fund_mapper.cik_to_class_ids
    assert len(cik_to_class_ids) == mutual_fund_mapper.raw_dataframe.CIK.nunique()

    class_id = vtsax_mutual_fund["classId"]
    cik = vtsax_mutual_fund["cik"]
    assert cik in cik_to_class_ids
    assert class_id in cik_to_class_ids[cik]


def test_class_id_to_cik(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    class_id_to_cik = mutual_fund_mapper.class_id_to_cik
    assert (
        len(class_id_to_cik) == mutual_fund_mapper.raw_dataframe["Class ID"].nunique()
    )

    class_id = vtsax_mutual_fund["classId"]
    cik = vtsax_mutual_fund["cik"]
    assert class_id in class_id_to_cik
    assert class_id_to_cik[class_id] == cik


def test_class_id_to_ticker(
    mutual_fund_mapper: MutualFundMapper, vtsax_mutual_fund: Dict[str, str]
):
    class_id_to_ticker = mutual_fund_mapper.class_id_to_ticker
    df = mutual_fund_mapper.raw_dataframe.Ticker
    assert len(class_id_to_ticker) == df[df != ""].nunique()

    class_id = vtsax_mutual_fund["classId"]
    ticker = vtsax_mutual_fund["ticker"]
    assert class_id in class_id_to_ticker
    assert class_id_to_ticker[class_id] == ticker


def test_caching(mutual_fund_mapper: MutualFundMapper):
    # Clear cache
    MutualFundMapper.ticker_to_cik.fget.cache_clear()  # type: ignore
    n = 1000

    for _ in range(n):
        mutual_fund_mapper.ticker_to_cik

    # Verify cache hits and misses
    cache_info = MutualFundMapper.ticker_to_cik.fget.cache_info()  # type: ignore

    expected_misses = 1
    assert cache_info.misses == expected_misses

    expected_hits = n - expected_misses
    assert cache_info.hits == expected_hits



from ._version import __version__
from .BaseMapper import BaseMapper
from .MutualFundMapper import MutualFundMapper
from .retrievers import BaseRetriever, MutualFundRetriever, StockRetriever
from .StockMapper import StockMapper



# mypy: ignore-errors

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

import os
import sys
from pathlib import Path

sys.path.insert(0, os.path.abspath(".."))

version = Path("../sec_cik_mapper/_version.py").read_text(encoding="utf-8")
about = {}
exec(version, about)

# -- Project information -----------------------------------------------------

project = "SEC CIK Mapper 🗺"
copyright = "2022, Jad Chaar"
author = "Jad Chaar"

release = about["__version__"]

# -- General configuration ---------------------------------------------------

extensions = ["sphinx.ext.autodoc", "sphinx_autodoc_typehints", "m2r2"]

autodoc_member_order = "groupwise"

master_doc = "index"
source_suffix = [".rst", ".md"]

html_sidebars = {
    "**": ["about.html", "localtoc.html", "relations.html", "searchbox.html"]
}

templates_path = ["_templates"]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

html_theme = "alabaster"
html_theme_options = {
    "description": "Generate mappings between SEC identifiers using Python",
    "github_user": "jadchaar",
    "github_repo": "sec-cik-mapper",
    "github_banner": True,
    "show_related": False,
    "show_powered_by": False,
}

html_static_path = []

html_show_sourcelink = False
html_show_copyright = True
html_show_sphinx = False



import json
from pathlib import Path

import pandas as pd

from sec_cik_mapper import MutualFundMapper, StockMapper

NUM_STOCK_FILES = 9
NUM_CSV_FILES = 1
NUM_MUTUAL_FUND_FILES = 12


def test_validate_num_files_stocks(
    stock_mapper: StockMapper, generated_mappings_path_stocks: Path
):
    all_files = list(generated_mappings_path_stocks.glob("*"))
    assert len(all_files) == NUM_STOCK_FILES

    csv_files = list(generated_mappings_path_stocks.glob("*.csv"))
    assert len(csv_files) == NUM_CSV_FILES

    json_files = list(generated_mappings_path_stocks.glob("*.json"))
    assert len(json_files) == NUM_STOCK_FILES - NUM_CSV_FILES


def test_validate_num_files_mutual_funds(
    mutual_fund_mapper: MutualFundMapper, generated_mappings_path_mutual_funds: Path
):
    all_files = list(generated_mappings_path_mutual_funds.glob("*"))
    assert len(all_files) == NUM_MUTUAL_FUND_FILES

    csv_files = list(generated_mappings_path_mutual_funds.glob("*.csv"))
    assert len(csv_files) == NUM_CSV_FILES

    json_files = list(generated_mappings_path_mutual_funds.glob("*.json"))
    assert len(json_files) == NUM_MUTUAL_FUND_FILES - NUM_CSV_FILES


def test_validate_csv_stocks(
    stock_mapper: StockMapper, generated_mappings_path_stocks: Path
):
    csv_path = generated_mappings_path_stocks / "mappings.csv"
    assert csv_path.exists()

    df = pd.read_csv(csv_path)
    assert len(df) == len(stock_mapper.raw_dataframe)


def test_validate_csv_mutual_funds(
    mutual_fund_mapper: MutualFundMapper, generated_mappings_path_mutual_funds: Path
):
    csv_path = generated_mappings_path_mutual_funds / "mappings.csv"
    assert csv_path.exists()

    df = pd.read_csv(csv_path)
    assert len(df) == len(mutual_fund_mapper.raw_dataframe)


def test_validate_json(
    generated_mappings_path_stocks: Path, generated_mappings_path_mutual_funds: Path
):
    generated_mappings_paths = list(
        generated_mappings_path_stocks.glob("*.json")
    ) + list(generated_mappings_path_mutual_funds.glob("*.json"))
    assert (
        len(generated_mappings_paths)
        == NUM_MUTUAL_FUND_FILES + NUM_STOCK_FILES - 2 * NUM_CSV_FILES
    )

    for json_path in generated_mappings_paths:
        with json_path.open() as f:
            json_obj = json.load(f)
        assert len(json_obj) > 0



from pathlib import Path
from typing import Dict

import pandas as pd

from sec_cik_mapper import StockMapper


def test_stock_mapper_initialization(stock_mapper: StockMapper):
    df = stock_mapper.raw_dataframe

    # Verify dataframe rows
    assert len(df) > 0

    # Verify dataframe columns
    df_columns = list(df.columns)
    assert len(df_columns) == 4
    assert "CIK" in df_columns
    assert "Ticker" in df_columns
    assert "Name" in df_columns
    assert "Exchange" in df_columns

    validate_dataframe_content(df)


def validate_dataframe_content(df: pd.DataFrame):
    for cik, ticker, company, exchange in zip(
        df["CIK"], df["Ticker"], df["Name"], df["Exchange"]
    ):
        assert isinstance(cik, str) and len(cik) == 10
        assert isinstance(ticker, str) and len(ticker) > 0
        assert isinstance(company, str) and len(company) > 0
        # SEC outputs blank exchanges for some stocks
        assert isinstance(exchange, str)


def test_cik_to_tickers(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    cik_to_tickers = stock_mapper.cik_to_tickers
    assert len(cik_to_tickers) == stock_mapper.raw_dataframe.CIK.nunique()

    # Deal with CIKs mapping to multiple tickers
    num_tickers = 0
    for cik, tickers in cik_to_tickers.items():
        assert isinstance(tickers, set)
        num_tickers += len(tickers)
        assert len(cik) == 10

    assert num_tickers == len(stock_mapper.raw_dataframe.CIK)

    ticker = apple_stock["ticker"]
    cik = apple_stock["cik"]
    assert cik in cik_to_tickers
    assert ticker in cik_to_tickers[cik]


def test_ticker_to_cik(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    ticker_to_cik = stock_mapper.ticker_to_cik
    df = stock_mapper.raw_dataframe.Ticker
    assert len(ticker_to_cik) == df[df != ""].nunique()

    ticker = apple_stock["ticker"]
    cik = apple_stock["cik"]
    assert ticker in ticker_to_cik
    assert len(ticker_to_cik[ticker]) == 10
    assert ticker_to_cik[ticker] == cik


def test_cik_to_company_name(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    cik_to_company_name = stock_mapper.cik_to_company_name
    assert len(cik_to_company_name) == stock_mapper.raw_dataframe.CIK.nunique()

    cik = apple_stock["cik"]
    name = apple_stock["name"]
    assert cik in cik_to_company_name
    assert cik_to_company_name[cik] == name


def test_ticker_to_company_name(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    ticker_to_company_name = stock_mapper.ticker_to_company_name
    df = stock_mapper.raw_dataframe.Ticker
    assert len(ticker_to_company_name) == df[df != ""].nunique()

    ticker = apple_stock["ticker"]
    name = apple_stock["name"]
    assert ticker in ticker_to_company_name
    assert ticker_to_company_name[ticker] == name


def test_ticker_to_exchange(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    ticker_to_exchange = stock_mapper.ticker_to_exchange
    df = stock_mapper.raw_dataframe
    assert (
        len(ticker_to_exchange)
        == df[(df.Ticker != "") & (df.Exchange != "")].Ticker.nunique()
    )

    ticker = apple_stock["ticker"]
    exchange = apple_stock["exchange"]
    assert ticker in ticker_to_exchange
    assert ticker_to_exchange[ticker] == exchange


def test_exchange_to_tickers(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    exchange_to_tickers = stock_mapper.exchange_to_tickers
    exchange_series = stock_mapper.raw_dataframe["Exchange"]
    assert len(exchange_to_tickers) == exchange_series[exchange_series != ""].nunique()

    ticker = apple_stock["ticker"]
    exchange = apple_stock["exchange"]
    assert exchange in exchange_to_tickers
    assert isinstance(exchange_to_tickers[exchange], set)
    assert ticker in exchange_to_tickers[exchange]


def test_cik_to_exchange(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    cik_to_exchange = stock_mapper.cik_to_exchange
    df = stock_mapper.raw_dataframe
    assert len(cik_to_exchange) == df[df["Exchange"] != ""]["CIK"].nunique()

    cik = apple_stock["cik"]
    exchange = apple_stock["exchange"]
    assert cik in cik_to_exchange
    assert cik_to_exchange[cik] == exchange


def test_exchange_to_ciks(stock_mapper: StockMapper, apple_stock: Dict[str, str]):
    exchange_to_ciks = stock_mapper.exchange_to_ciks
    df = stock_mapper.raw_dataframe
    assert len(exchange_to_ciks) == df[df["Exchange"] != ""]["Exchange"].nunique()

    cik = apple_stock["cik"]
    exchange = apple_stock["exchange"]
    assert exchange in exchange_to_ciks
    assert cik in exchange_to_ciks[exchange]


def test_raw_dataframe(stock_mapper: StockMapper):
    assert stock_mapper.raw_dataframe.equals(stock_mapper.mapping_metadata)


def test_save_metadata_to_csv(stock_mapper: StockMapper, tmp_path: Path):
    tmp_csv_path = tmp_path / "temp.csv"
    stock_mapper.save_metadata_to_csv(tmp_csv_path)
    assert tmp_csv_path.exists()
    df = pd.read_csv(tmp_csv_path)
    assert len(df) == len(stock_mapper.raw_dataframe)


def test_caching(stock_mapper: StockMapper):
    # Clear cache
    StockMapper.ticker_to_cik.fget.cache_clear()  # type: ignore
    n = 1000

    for _ in range(n):
        stock_mapper.ticker_to_cik

    # Verify cache hits and misses
    cache_info = StockMapper.ticker_to_cik.fget.cache_info()  # type: ignore

    expected_misses = 1
    assert cache_info.misses == expected_misses

    expected_hits = n - expected_misses
    assert cache_info.hits == expected_hits



from pathlib import Path
from typing import Dict

import pytest

from sec_cik_mapper import (
    MutualFundMapper,
    MutualFundRetriever,
    StockMapper,
    StockRetriever,
)


@pytest.fixture(scope="session")
def stock_mapper() -> StockMapper:
    return StockMapper()


@pytest.fixture(scope="session")
def mutual_fund_mapper() -> MutualFundMapper:
    return MutualFundMapper()


@pytest.fixture(scope="session")
def generated_mappings_path_stocks() -> Path:
    return Path("mappings/stocks")


@pytest.fixture(scope="session")
def generated_mappings_path_mutual_funds() -> Path:
    return Path("mappings/mutual_funds")


@pytest.fixture(scope="session")
def stock_retriever() -> StockRetriever:
    return StockRetriever()


@pytest.fixture(scope="session")
def mutual_fund_retriever() -> MutualFundRetriever:
    return MutualFundRetriever()


@pytest.fixture(scope="session")
def apple_stock() -> Dict[str, str]:
    return {
        "ticker": "AAPL",
        "cik": "0000320193",
        "name": "Apple Inc.",
        "exchange": "Nasdaq",
    }


@pytest.fixture(scope="session")
def vtsax_mutual_fund() -> Dict[str, str]:
    return {
        "ticker": "VTSAX",
        "cik": "0000036405",
        "seriesId": "S000002848",
        "classId": "C000007806",
    }



"""Provides a :class:`StockMapper` class for mapping CIKs, tickers,
exchanges, and company names."""

from typing import ClassVar, Dict

from .BaseMapper import BaseMapper
from .retrievers import StockRetriever
from .types import KeyToValueSet
from .utils import with_cache


class StockMapper(BaseMapper):
    """A :class:`StockMapper` object.

    Usage::

        >>> from sec_cik_mapper import StockMapper
        >>> stock_mapper = StockMapper()
    """

    _retriever: ClassVar[StockRetriever] = StockRetriever()

    def __init__(self) -> None:
        """Constructor for the :class:`StockMapper` class."""
        super().__init__(StockMapper._retriever)

    @property  # type: ignore
    @with_cache
    def cik_to_company_name(self) -> Dict[str, str]:
        """Get CIK to company name mapping.

        Usage::

            >>> from sec_cik_mapper import StockMapper
            >>> stock_mapper = StockMapper()
            >>> stock_mapper.cik_to_company_name
            {'0000320193': 'Apple Inc.', '0000789019': 'Microsoft Corp', ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        company_name_col = self.mapping_metadata["Name"]
        return self._form_kv_mapping(cik_col, company_name_col)

    @property  # type: ignore
    @with_cache
    def ticker_to_company_name(self) -> Dict[str, str]:
        """Get ticker to company name mapping.

        Usage::

            >>> from sec_cik_mapper import StockMapper
            >>> stock_mapper = StockMapper()
            >>> stock_mapper.ticker_to_company_name
            {'AAPL': 'Apple Inc.', 'MSFT': 'Microsoft Corp', 'GOOG': 'Alphabet Inc.', ...}
        """
        ticker_col = self.mapping_metadata["Ticker"]
        company_name_col = self.mapping_metadata["Name"]
        return self._form_kv_mapping(ticker_col, company_name_col)

    @property  # type: ignore
    @with_cache
    def ticker_to_exchange(self) -> Dict[str, str]:
        """Get ticker to exchange mapping.

        Usage::

            >>> from sec_cik_mapper import StockMapper
            >>> stock_mapper = StockMapper()
            >>> stock_mapper.ticker_to_exchange
            {'AAPL': 'Nasdaq', 'MSFT': 'Nasdaq', 'GOOG': 'Nasdaq', ...}
        """
        ticker_col = self.mapping_metadata["Ticker"]
        exchange_col = self.mapping_metadata["Exchange"]
        return self._form_kv_mapping(ticker_col, exchange_col)

    @property  # type: ignore
    @with_cache
    def exchange_to_tickers(self) -> KeyToValueSet:
        """Get exchange to tickers mapping.

        Usage::

            >>> from sec_cik_mapper import StockMapper
            >>> stock_mapper = StockMapper()
            >>> stock_mapper.exchange_to_tickers
            {'Nasdaq': {'CYRN', 'OHPAW', ...}, 'NYSE': {'PLAG', 'TDW-WTB', ...}, ...}
        """
        ticker_col = self.mapping_metadata["Ticker"]
        exchange_col = self.mapping_metadata["Exchange"]
        return self._form_kv_set_mapping(exchange_col, ticker_col)

    @property  # type: ignore
    @with_cache
    def cik_to_exchange(self) -> Dict[str, str]:
        """Get CIK to exchange mapping.

        Usage::

            >>> from sec_cik_mapper import StockMapper
            >>> stock_mapper = StockMapper()
            >>> stock_mapper.cik_to_exchange
            {'0000320193': 'Nasdaq', '0000789019': 'Nasdaq', '0001652044': 'Nasdaq', ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        exchange_col = self.mapping_metadata["Exchange"]
        return self._form_kv_mapping(cik_col, exchange_col)

    @property  # type: ignore
    @with_cache
    def exchange_to_ciks(self) -> KeyToValueSet:
        """Get exchange to CIKs mapping.

        Usage::

            >>> from sec_cik_mapper import StockMapper
            >>> stock_mapper = StockMapper()
            >>> stock_mapper.exchange_to_ciks
            {'Nasdaq': {'0000779544', ...}, 'NYSE': {'0000764478', ...}, ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        exchange_col = self.mapping_metadata["Exchange"]
        return self._form_kv_set_mapping(exchange_col, cik_col)



from typing import List, Union

import pytest

from sec_cik_mapper import BaseRetriever, MutualFundRetriever, StockRetriever, types


def test_base_retriever_instantiation_type_error():
    # BaseRetriever cannot be directly instantiated
    with pytest.raises(TypeError):
        _ = BaseRetriever()


def test_retriever_source_urls(
    stock_retriever: StockRetriever, mutual_fund_retriever: MutualFundRetriever
):
    assert len(stock_retriever.source_url) > 0
    assert len(mutual_fund_retriever.source_url) > 0
    assert stock_retriever.source_url != mutual_fund_retriever.source_url


def test_clean_ticker(stock_retriever: StockRetriever):
    # Remove non-alphanumeric characters (except dash)
    assert stock_retriever._clean_ticker("(NWAGX)") == "NWAGX"

    # Keep dashes
    assert stock_retriever._clean_ticker("BRK-A") == "BRK-A"

    # Keep numeric characters
    assert stock_retriever._clean_ticker("FAKE8") == "FAKE8"

    # Make all caps
    assert stock_retriever._clean_ticker("tIcKeR") == "TICKER"


def test_stock_retriever_transform(stock_retriever: StockRetriever):
    field_indices: types.FieldIndices = {
        "cik": 0,
        "name": 1,
        "ticker": 2,
        "exchange": 3,
    }
    company_data: List[Union[int, str]] = [320193, "Apple Inc.", "AAPL", "Nasdaq"]
    expected = {
        "CIK": "0000320193",
        "Ticker": "AAPL",
        "Name": "Apple Inc.",
        "Exchange": "Nasdaq",
    }
    assert stock_retriever.transform(field_indices, company_data) == expected


def test_mutual_fund_retriever_transform(mutual_fund_retriever: MutualFundRetriever):
    field_indices: types.FieldIndices = {
        "cik": 0,
        "seriesId": 1,
        "classId": 2,
        "symbol": 3,
    }
    company_data: List[Union[int, str]] = [2110, "S000009184", "C000024954", "LACAX"]
    expected = {
        "CIK": "0000002110",
        "Ticker": "LACAX",
        "Series ID": "S000009184",
        "Class ID": "C000024954",
    }
    assert mutual_fund_retriever.transform(field_indices, company_data) == expected



import pytest

from sec_cik_mapper import BaseMapper


def test_base_mapper_instantiation_type_error():
    # BaseMapper cannot be directly instantiated
    with pytest.raises(TypeError):
        _ = BaseMapper()



from typing import Dict, List, Set, TypeVar, Union

from typing_extensions import Literal, TypedDict


class StockFieldIndices(TypedDict):
    cik: int
    ticker: int
    name: int
    exchange: int


class MutualFundFieldIndices(TypedDict):
    cik: int
    symbol: int
    seriesId: int
    classId: int


FieldIndices = Union[StockFieldIndices, MutualFundFieldIndices]

StockFields = Literal["cik", "name", "ticker", "exchange"]

MutualFundFields = Literal["cik", "seriesId", "classId", "symbol"]

Fields = Union[StockFields, MutualFundFields]

CompanyData = List[List[Union[int, str]]]

KeyToValueSet = Dict[str, Set[str]]

T = TypeVar("T")



"""Retrieval abstractions for SEC source URLs and data transformers."""

import re
from abc import ABCMeta, abstractmethod
from typing import ClassVar, Dict, List, Pattern, Union, cast

from typing_extensions import Final

from .types import FieldIndices, MutualFundFieldIndices, StockFieldIndices

# See CIK, ticker, and exchange associations section of:
# https://www.sec.gov/os/accessing-edgar-data
_SEC_MAPPING_SOURCE_URL_STOCKS: Final[
    str
] = "https://www.sec.gov/files/company_tickers_exchange.json"
_SEC_MAPPING_SOURCE_URL_MUTUAL_FUNDS: Final[
    str
] = "https://www.sec.gov/files/company_tickers_mf.json"


class BaseRetriever(metaclass=ABCMeta):
    # Tickers can contain letters, numbers, and dashes
    ticker_pattern: ClassVar[Pattern[str]] = re.compile(r"[^A-Z0-9\-]+")

    def _clean_ticker(self, string: str) -> str:
        return re.sub(BaseRetriever.ticker_pattern, "", string.upper())

    @property
    @abstractmethod
    def source_url(self) -> str:
        """Get source URL for retriever."""

    @abstractmethod
    def transform(
        self,
        field_indices: FieldIndices,
        company_data: List[Union[int, str]],
    ) -> Dict[str, str]:
        """Transform company data depending on whether it is a stock or a mutual fund."""


class StockRetriever(BaseRetriever):
    @property
    def source_url(self) -> str:
        return _SEC_MAPPING_SOURCE_URL_STOCKS

    def transform(
        self,
        field_indices: FieldIndices,
        company_data: List[Union[int, str]],
    ) -> Dict[str, str]:
        field_indices = cast(StockFieldIndices, field_indices)
        cik = str(company_data[field_indices["cik"]])
        ticker = str(company_data[field_indices["ticker"]])
        name = str(company_data[field_indices["name"]])
        exchange = str(company_data[field_indices["exchange"]])
        return {
            "CIK": cik.zfill(10),
            "Ticker": self._clean_ticker(ticker),
            "Name": name.title(),
            "Exchange": exchange,
        }


class MutualFundRetriever(BaseRetriever):
    @property
    def source_url(self) -> str:
        return _SEC_MAPPING_SOURCE_URL_MUTUAL_FUNDS

    def transform(
        self,
        field_indices: FieldIndices,
        company_data: List[Union[int, str]],
    ) -> Dict[str, str]:
        field_indices = cast(MutualFundFieldIndices, field_indices)
        cik = str(company_data[field_indices["cik"]])
        seriesId = str(company_data[field_indices["seriesId"]])
        classId = str(company_data[field_indices["classId"]])
        symbol = str(company_data[field_indices["symbol"]])
        return {
            "CIK": cik.zfill(10),
            "Ticker": self._clean_ticker(symbol),
            "Series ID": seriesId,
            "Class ID": classId,
        }



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['sec_cik_mapper']

package_data = \
{'': ['*']}

install_requires = \
['requests', 'pandas', 'typing_extensions']

extras_require = \
{'doc': ['sphinx', 'sphinx-autodoc-typehints'],
 'test': ['pytest', 'pytest-cov']}

setup(name='sec-cik-mapper',
      version='2.1.0',
      description='Generate mappings between SEC identifiers using Python',
      author=None,
      author_email='Jad Chaar <jad.chaar@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



"""Provides a :class:`BaseMapper` class for mapping stock and mutual
fund data from the SEC."""

import time
from collections import defaultdict
from pathlib import Path
from typing import ClassVar, Dict, List, Union, cast

import pandas as pd
import requests

from .retrievers import MutualFundRetriever, StockRetriever
from .types import CompanyData, FieldIndices, Fields, KeyToValueSet
from .utils import with_cache


class BaseMapper:
    """A :class:`BaseMapper` object."""

    _headers: ClassVar[Dict[str, str]] = {
        "User-Agent": f"{int(time.time())} {int(time.time())}@gmail.com",
        "Accept-Encoding": "gzip, deflate",
        "Host": "www.sec.gov",
    }

    def __init__(self, retriever: Union[StockRetriever, MutualFundRetriever]) -> None:
        """Constructor for the :class:`BaseMapper` class."""
        self.retriever = retriever
        self.mapping_metadata = self._get_mapping_metadata_from_sec()

    def __new__(cls, *args, **kwargs):
        """BaseMapper should not be directly instantiated,
        so throw an error on instantiation.

        More info: https://stackoverflow.com/a/7990308/3820660
        """
        if cls is BaseMapper:
            raise TypeError(
                f"{cls.__name__} cannot be directly instantiated. "
                "Please instantiate the StockMapper and/or MutualFundMapper "
                "classes instead."
            )
        return object.__new__(cls, *args, **kwargs)

    def _get_indices_from_fields(self, fields: Fields) -> FieldIndices:
        """Get list indices from field names."""
        field_indices = {field: fields.index(field) for field in fields}
        return cast(FieldIndices, field_indices)

    def _get_mapping_metadata_from_sec(self) -> pd.DataFrame:
        """Get company mapping metadata from the SEC as a pandas dataframe,
        sorted by CIK and ticker.
        """
        resp = requests.get(self.retriever.source_url, headers=BaseMapper._headers)
        resp.raise_for_status()
        data = resp.json()

        fields: Fields = data["fields"]
        field_indices: FieldIndices = self._get_indices_from_fields(fields)

        company_data: CompanyData = data["data"]
        transformed_data: List[Dict[str, str]] = []

        for cd in company_data:
            transformed_data.append(
                self.retriever.transform(field_indices, cd),
            )

        df = pd.DataFrame(transformed_data)
        df.sort_values(by=["CIK", "Ticker"], inplace=True, ignore_index=True)
        return df

    def _form_kv_set_mapping(self, keys: pd.Series, values: pd.Series) -> KeyToValueSet:
        """Form mapping from key to list of values, ignoring blank keys and values.

        Example: numerous CIKs map to multiple tickers (e.g. Banco Santander),
        so we must keep a list of tickers for each unique CIK.
        """
        mapping = defaultdict(set)
        for key, value in zip(keys, values):
            # Ignore blank keys and values
            if key and value:
                mapping[key].add(value)
        return dict(mapping)

    def _form_kv_mapping(self, keys: pd.Series, values: pd.Series) -> Dict[str, str]:
        """Form key-value mapping, ignoring blank keys and values."""
        return {k: v for k, v in zip(keys, values) if k and v}

    @property  # type: ignore
    @with_cache
    def cik_to_tickers(self) -> KeyToValueSet:
        """Get CIK to tickers mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper, StockMapper
            >>> stock_mapper = StockMapper()
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> stock_mapper.cik_to_tickers
            {'0000320193': {'AAPL'}, '0001652044': {'GOOG', 'GOOGL'}, ...}
            >>> mutual_fund_mapper.cik_to_tickers
            {'0000002110': {'CRBYX', 'CEFZX', ...}, '0000002646': {'IIBPX', 'IPISX', ...}, ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        ticker_col = self.mapping_metadata["Ticker"]
        return self._form_kv_set_mapping(cik_col, ticker_col)

    @property  # type: ignore
    @with_cache
    def ticker_to_cik(self) -> Dict[str, str]:
        """Get ticker to CIK mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper, StockMapper
            >>> stock_mapper = StockMapper()
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> stock_mapper.ticker_to_cik
            {'AAPL': '0000320193', 'MSFT': '0000789019', 'GOOG': '0001652044', ...}
            >>> mutual_fund_mapper.ticker_to_cik
            {'LACAX': '0000002110', 'LIACX': '0000002110', 'ACRNX': '0000002110', ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        ticker_col = self.mapping_metadata["Ticker"]
        return self._form_kv_mapping(ticker_col, cik_col)

    @property  # type: ignore
    def raw_dataframe(self) -> pd.DataFrame:
        """Get raw pandas dataframe.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper, StockMapper
            >>> stock_mapper = StockMapper()
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> stock_mapper.raw_dataframe
                        CIK  Ticker                                  Name Exchange
            0      0000320193    AAPL                            Apple Inc.   Nasdaq
            1      0000789019    MSFT                        Microsoft Corp   Nasdaq
            2      0001652044    GOOG                         Alphabet Inc.   Nasdaq
            3      0001018724    AMZN                        Amazon Com Inc   Nasdaq
            4      0001318605    TSLA                           Tesla, Inc.   Nasdaq
            ...           ...     ...                                   ...      ...
            13184  0001866816   OLITU             Omnilit Acquisition Corp.   Nasdaq
            13185  0001870778   OHAAU               Opy Acquisition Corp. I   Nasdaq
            13186  0001873324   PEPLW    Pepperlime Health Acquisition Corp   Nasdaq
            13187  0001877557  WEL-UN  Integrated Wellness Acquisition Corp     NYSE
            13188  0001877787  ZGN-WT   Ermenegildo Zegna Holditalia S.P.A.     NYSE

            [13189 rows x 4 columns]
            >>> mutual_fund_mapper.raw_dataframe
                            CIK Ticker   Series ID    Class ID
            0      0000002110  LACAX  S000009184  C000024954
            1      0000002110  LIACX  S000009184  C000024956
            2      0000002110  ACRNX  S000009184  C000024957
            3      0000002110  CEARX  S000009184  C000122735
            4      0000002110  CRBRX  S000009184  C000122736
            ...           ...    ...         ...         ...
            29237  0001860434   SIHY  S000072555  C000228888
            29238  0001860434   SIFI  S000072556  C000228889
            29239  0001860434   INNO  S000073580  C000230585
            29240  0001877493    BTF  S000074058  C000231452
            29241  0001877493    VBB  S000075054  C000233857

            [29242 rows x 4 columns]
        """
        return self.mapping_metadata

    def save_metadata_to_csv(self, path: Union[str, Path]) -> None:
        """Save stock mapping metadata (CIK, ticker, exchange, and company name)
        or mutual fund mapping metadata (CIK, ticker, series ID, class ID) from
        SEC to CSV.

        Usage::

            >>> from sec_cik_mapper import StockMapper, MutualFundMapper
            >>> from pathlib import Path
            >>> stock_mapper = StockMapper()
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> csv_path = Path("cik_mapping.csv")
            # Save full CIK, ticker, exchange, and company name mapping to a CSV file
            >>> stock_mapper.save_metadata_to_csv(csv_path)
            # Save full CIK, ticker, series ID, and class ID mapping to a CSV file
            >>> mutual_fund_mapper.save_metadata_to_csv(csv_path)
        """
        self.mapping_metadata.to_csv(path, index=False)



"""Provides a :class:`MutualFundMapper` class for mapping CIKs, tickers,
series IDs, and class IDs."""

from typing import ClassVar, Dict

from .BaseMapper import BaseMapper
from .retrievers import MutualFundRetriever
from .types import KeyToValueSet
from .utils import with_cache


class MutualFundMapper(BaseMapper):
    """A :class:`MutualFundMapper` object.

    Usage::

        >>> from sec_cik_mapper import MutualFundMapper
        >>> mutual_fund_mapper = MutualFundMapper()
    """

    _retriever: ClassVar[MutualFundRetriever] = MutualFundRetriever()

    def __init__(self) -> None:
        """Constructor for the :class:`MutualFundMapper` class."""
        super().__init__(MutualFundMapper._retriever)

    @property  # type: ignore
    @with_cache
    def cik_to_series_ids(self) -> KeyToValueSet:
        """Get CIK to series ID mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.cik_to_series_ids
            {'0000002110': {'S000009184', 'S000033622', ...}, '0000002646': {'S000008760'}, ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        series_id_col = self.mapping_metadata["Series ID"]
        return self._form_kv_set_mapping(cik_col, series_id_col)

    @property  # type: ignore
    @with_cache
    def ticker_to_series_id(self) -> Dict[str, str]:
        """Get ticker to series ID mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.ticker_to_series_id
            {'LACAX': 'S000009184', 'LIACX': 'S000009184', 'ACRNX': 'S000009184', ...}
        """
        ticker_col = self.mapping_metadata["Ticker"]
        series_id_col = self.mapping_metadata["Series ID"]
        return self._form_kv_mapping(ticker_col, series_id_col)

    @property  # type: ignore
    @with_cache
    def series_id_to_cik(self) -> Dict[str, str]:
        """Get series ID to CIK mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.series_id_to_cik
            {'S000009184': '0000002110', 'S000009185': '0000002110', ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        series_id_col = self.mapping_metadata["Series ID"]
        return self._form_kv_mapping(series_id_col, cik_col)

    @property  # type: ignore
    @with_cache
    def series_id_to_tickers(self) -> KeyToValueSet:
        """Get series ID to tickers mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.series_id_to_tickers
            {'S000009184': {'CEARX', 'CRBYX', ...}, 'S000009185': {'ACINX', 'CACRX', ...}, ...}
        """
        ticker_col = self.mapping_metadata["Ticker"]
        series_id_col = self.mapping_metadata["Series ID"]
        return self._form_kv_set_mapping(series_id_col, ticker_col)

    @property  # type: ignore
    @with_cache
    def series_id_to_class_ids(self) -> KeyToValueSet:
        """Get series ID to class IDs mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.series_id_to_class_ids
            {'S000009184': {'C000024956', ...}, 'S000009185': {'C000024958', ...}, ...}
        """
        class_id_col = self.mapping_metadata["Class ID"]
        series_id_col = self.mapping_metadata["Series ID"]
        return self._form_kv_set_mapping(series_id_col, class_id_col)

    @property  # type: ignore
    @with_cache
    def ticker_to_class_id(self) -> Dict[str, str]:
        """Get ticker to class ID mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.ticker_to_class_id
            {'LACAX': 'C000024954', 'LIACX': 'C000024956', 'ACRNX': 'C000024957', ...}
        """
        ticker_col = self.mapping_metadata["Ticker"]
        class_id_col = self.mapping_metadata["Class ID"]
        return self._form_kv_mapping(ticker_col, class_id_col)

    @property  # type: ignore
    @with_cache
    def cik_to_class_ids(self) -> KeyToValueSet:
        """Get CIK to class IDs mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.cik_to_class_ids
            {'0000002110': {'C000024958', ...}, '0000002646': {'C000023849', ...}, ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        class_id_col = self.mapping_metadata["Class ID"]
        return self._form_kv_set_mapping(cik_col, class_id_col)

    @property  # type: ignore
    @with_cache
    def class_id_to_cik(self) -> Dict[str, str]:
        """Get class ID to CIK mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.class_id_to_cik
            {'C000024954': '0000002110', 'C000024956': '0000002110', ...}
        """
        cik_col = self.mapping_metadata["CIK"]
        class_id_col = self.mapping_metadata["Class ID"]
        return self._form_kv_mapping(class_id_col, cik_col)

    @property  # type: ignore
    @with_cache
    def class_id_to_ticker(self) -> Dict[str, str]:
        """Get class ID to ticker mapping.

        Usage::

            >>> from sec_cik_mapper import MutualFundMapper
            >>> mutual_fund_mapper = MutualFundMapper()
            >>> mutual_fund_mapper.class_id_to_ticker
            {'C000024954': 'LACAX', 'C000024956': 'LIACX', 'C000024957': 'ACRNX', ...}
        """
        ticker_col = self.mapping_metadata["Ticker"]
        class_id_col = self.mapping_metadata["Class ID"]
        return self._form_kv_mapping(class_id_col, ticker_col)
